@@ -1,10 +1,15 @@
 import json
 from contextlib import contextmanager
+from typing import Union
 
 from tokenstream import TokenStream
 from tokenstream.error import InvalidSyntax
 
-from .json_parser import parse_json
+from allay.json_parser import parse_json
+
+
+class DefinitionAlreadyExists(Exception):
+    ...
 
 
 class Parser:
@@ -12,44 +17,67 @@ class Parser:
     Parser - The base class for all allay parsers
 
     Args:
+        indent (int, optional): Indentation level. Ignored if ``json_dump`` is False. Defaults to None.
+        json_dump (bool, optional): Whether to dump the output as JSON in a string. Defaults to None.
         definition_delimeter (str, optional): The delimeter that should separate the definitions from the text. Defaults to "#ALLAYDEFS\n".
     """
 
-    def __init__(self, definition_delimeter: str = "#ALLAYDEFS\n") -> None:
+    def __init__(
+        self,
+        indent: int = None,
+        json_dump: bool = None,
+        definition_delimeter: str = "#ALLAYDEFS\n",
+    ) -> None:
+        # Set default parser options
+        self.indent = indent
+        self.json_dump = json_dump
+        self.set_defaults(indent, json_dump)
+
         self.definition_delimeter = definition_delimeter
 
         self.patterns = {}
         self.templates = {}
+        self.parent = ""
 
     def parse(
-        self, input: str, indent: int = None, json_dump: bool = True
-    ) -> str | dict:
+        self, text: str, indent: int = None, json_dump: bool = None
+    ) -> Union[str, dict]:
         """
         parse - Converts a string into a text-component using the Allay format
 
         Args:
-            input (str): The string to parse, or a file path
-            indent (int, optional): Indentation level. Defaults to None. Ignored if ``json_dump`` is False.
-            json_dump (bool, optional): Whether to dump the output as JSON in a string. Defaults to True.
+            text (str): The text to parse, or a file path
+            indent (int, optional): Indentation level. Ignored if ``json_dump`` is False. Defaults to None.
+            json_dump (bool, optional): Whether to dump the output as JSON in a string. Defaults to None.
 
         Raises:
             InvalidSyntax: The syntax is invalid
 
         Returns:
-            str: The text-component
+            Union[str, dict]: The text-component
         """
+        # Grab defaults (if neither defaults nor function-level config is set, then it should remain None and work fine)
+        if not indent and self.indent:
+            indent = self.indent
+
+        if json_dump is None:
+            json_dump = self.json_dump if self.json_dump is not None else True
+
         # Check if it's a file, and if it is, use the file contents as the input
         try:
-            file = f'File "{input}"'  # File path
-            with open(input, "r") as infile:
-                input = infile.read()
-        except (FileNotFoundError, OSError) as error:
+            file = f'File "{text}"'  # File path
+            with open(text, "r") as infile:
+                text = infile.read()
+        except OSError as error:
             file = '"src"'
 
         # Print errors but with pizzaz
         try:
-            stream = TokenStream(self.pre_process(input))
+            stream = TokenStream(self.pre_process(text))
             output = self.internal_parse(stream)
+
+            self.parent = ""
+
             if json_dump:
                 return json.dumps(output, indent=indent)
             return output
@@ -63,6 +91,43 @@ class Parser:
                 )
             )
 
+    def set_defaults(self, indent: int = None, json_dump: bool = None) -> None:
+        """
+        set_defaults - Sets the default values for the parse function
+
+        Args:
+            indent (int, optional): Indentation level. Ignored if ``json_dump`` is False. Defaults to None.
+            json_dump (bool, optional): Whether to dump the output as JSON in a string. Defaults to None.
+        """
+        if indent:
+            self.indent = indent
+
+        if json_dump:
+            self.json_dump = json_dump
+
+    def set_parent(
+        self, stream: TokenStream, auto_generate_stream: bool = True
+    ) -> dict:
+        if auto_generate_stream and not isinstance(stream, TokenStream):
+            stream = TokenStream(stream)
+
+        with stream.syntax(paren=r"\(|\)", null=r"null|NULL"):
+            if stream.get(("paren", "(")):
+                with self.primary_syntax_definitions(stream):
+                    contents = self.parse_modifiers(stream)
+                stream.expect(("paren", ")"))
+
+                self.parent = contents
+                return contents
+
+            elif stream.get("null"):
+                self.parent = ""
+                return ""
+            else:
+                raise InvalidSyntax(
+                    "Expected either modifier or null for parent, got neither"
+                )
+
     def pre_process(self, text: str) -> str:
         # Separate the special components (patterns and templates)
         defs_end_index = text.find(self.definition_delimeter)
@@ -77,6 +142,9 @@ class Parser:
 
         return text.strip()
 
+    def get_string(self, stream: Union[str, TokenStream]) -> str:
+        return stream.expect("string").value[1:-1].replace('\\"', '"')
+
     @contextmanager
     def primary_syntax_definitions(self, stream: TokenStream):
         with stream.syntax(
@@ -96,7 +164,7 @@ class Parser:
             keybind=r"advancements|attack|back|chat|command|drop|forward|fullscreen|hotbar|inventory|jump|left|loadToolbarActivator|pickItem|playerlist|right|saveToolbarActivator|screenshot|smoothCamera|sneak|socialInteractions|spectatorOutlines|sprint|swapOffhand|togglePerspective|use",
             selector=r"@[parse](\[.*\])?",
             boolean=r"true|false",
-            string=r"\"(?:\\.|[^\\\n])*?\"",
+            string=r'"(?:\\.|[^"\\])*"',
             # Keywords sorted by type
             kw_json=r"hover_item",
             kw_scope=r"hover_text",
@@ -113,32 +181,36 @@ class Parser:
         self,
         type: str,
         name: str,
-        stream: str | TokenStream,
+        stream: Union[str, TokenStream],
         auto_generate_stream: bool,
     ) -> tuple:
         obj = self.patterns if type == "pattern" else self.templates
 
-        if name in obj:
-            raise InvalidSyntax(
+        name_prefix = "@" if type == "pattern" else "$"
+        internal_name = name if name.startswith(name_prefix) else name_prefix + name
+
+        if internal_name in obj:
+            raise DefinitionAlreadyExists(
                 f"{type.capitalize()} '{name}' has already been defined"
             )
 
-        if auto_generate_stream:
+        if auto_generate_stream and not isinstance(stream, TokenStream):
             stream = TokenStream(stream)
 
-        name_prefix = "@" if type == "pattern" else "$"
-
-        return stream, name_prefix + name if not name.startswith(name_prefix) else name
+        return stream, internal_name
 
     def add_pattern(
-        self, name: str, stream: str | TokenStream, auto_generate_stream: bool = True
-    ) -> str | dict:
+        self,
+        name: str,
+        stream: Union[str, TokenStream],
+        auto_generate_stream: bool = True,
+    ) -> Union[str, dict]:
         """
         add_pattern - Adds a pattern to the parser.
 
         Args:
             name (str): The name of the template. The beginning ``@`` is optional.
-            stream (str): The stream that contains the pattern. Make sure to enclose it in parenthesis. It can be a string or a ``TokenStream`` object. If a string, it will be converted to a ``TokenStream``.
+            stream (Union[str, TokenStream]): The stream that contains the pattern. Make sure to enclose it in parenthesis. It can be a string or a ``TokenStream`` object. If a string, it will be converted to a ``TokenStream``.
             auto_generate_stream (bool, optional): Whether or not to wrap ``stream`` in a ``TokenStream`` object. Only set this to ``False`` if you're converting it manually. Defaults to True.
 
         Returns:
@@ -148,7 +220,9 @@ class Parser:
             "pattern", name, stream, auto_generate_stream
         )
 
+        # Despite being defined in the outer scope, this token needs to be defined in here also because parenthesis are required when using add_pattern directly. Without this token definition, calling this function directly will result in expecting a parenthesis and fail.
         with stream.syntax(paren=r"\(|\)"):
+            # Parenthesis should be required in both add_pattern and in-text usage
             stream.expect(("paren", "("))
             with self.primary_syntax_definitions(stream):
                 contents = self.parse_modifiers(stream)
@@ -159,14 +233,17 @@ class Parser:
         return contents
 
     def add_template(
-        self, name: str, stream: str | TokenStream, auto_generate_stream: bool = True
+        self,
+        name: str,
+        stream: Union[str, TokenStream],
+        auto_generate_stream: bool = True,
     ) -> dict:
         """
         add_template - Adds a template to the parser.
 
         Args:
             name (str): The name of the template. The beginning ``$`` is optional.
-            stream (str): The stream that contains the template. Do not include the surrounding braces. It can be a string or a ``TokenStream`` object. If a string, it will be converted to a ``TokenStream``.
+            stream (Union[str, TokenStream]): The stream that contains the template. Do not include the surrounding braces. It can be a string or a ``TokenStream`` object. If a string, it will be converted to a ``TokenStream``.
             auto_generate_stream (bool, optional): Whether or not to wrap ``stream`` in a ``TokenStream`` object. Only set this to ``False`` if you're converting it manually. Defaults to True.
 
         Returns:
@@ -186,22 +263,31 @@ class Parser:
         with stream.syntax(
             pattern=r"@\w+",
             template=r"\$\w+",
+            parent=r"PARENT",
             equals=r"=",
             paren=r"\(|\)",
             brace=r"{|}",
         ), stream.intercept("newline"):
 
             try:
-                pattern, template = stream.expect("pattern", "template")
+                pattern, template, parent = stream.expect(
+                    "pattern", "template", "parent"
+                )
                 stream.expect("equals")
 
                 if pattern:
-                    self.add_pattern(stream, pattern.value, auto_generate_stream=False)
+                    self.add_pattern(pattern.value, stream, auto_generate_stream=False)
 
                 elif template:
+                    # Expected here and not in add_template because braces should NOT be added when add_template is used directly
+                    stream.expect(("brace", "{"))
                     self.add_template(
-                        stream, template.value, auto_generate_stream=False
+                        template.value, stream, auto_generate_stream=False
                     )
+                    stream.expect(("brace", "}"))
+
+                elif parent:
+                    self.set_parent(stream, auto_generate_stream=False)
 
                 if stream.get("newline"):
                     # Handle new lines in-between arguments
@@ -212,9 +298,7 @@ class Parser:
 
             except InvalidSyntax as error:
                 # Check if there's no data left to parse. This handles there being extra newlines between any arguments and the #ALLAYDEFS keyword. If there's still data left, then there's a syntax error.
-                if stream.source[
-                    stream.current.location.pos : len(stream.source)
-                ].strip():
+                if stream.source[stream.current.location.pos :].strip():
                     raise error
 
     def internal_parse(self, stream: TokenStream) -> list:
@@ -225,7 +309,25 @@ class Parser:
                 ("sqrbr", "["), ("brace", "{"), "escape", "text"
             ):
                 if escape:
-                    output.append(escape.value[1])
+                    if escape.value[1] in {
+                        "b",
+                        "f",
+                        "n",
+                        "r",
+                        "t",
+                        "\\",
+                        "[",
+                        "]",
+                        "(",
+                        ")",
+                        "{",
+                        "}",
+                        "<",
+                        ">",
+                    }:
+                        output.append(escape.value[1])
+                    else:
+                        output.append(escape.value)
 
                 elif sqrbr:
                     modified_text = self.internal_parse(stream)
@@ -254,11 +356,6 @@ class Parser:
             output = self.convert_ast_to_json(output)
             return output
 
-    def parse_separator(self, stream: TokenStream) -> str:
-        stream.expect(("kw_standalone", "sep"), ("kw_standalone", "separator"))
-        stream.expect("equals")  # equal sign
-        return stream.expect("string").value[1:-1]
-
     def error_if_scope(self, stream: TokenStream) -> None:
         if stream.data.get("scoped"):
             raise InvalidSyntax("Unexpected scope")
@@ -276,9 +373,7 @@ class Parser:
             elif kw_standalone:
                 if kw_standalone.value in {"sep", "separator"}:
                     stream.expect("equals")
-                    standalone_contents["separator"] = stream.expect("string").value[
-                        1:-1
-                    ]
+                    standalone_contents["separator"] = self.get_string(stream)
 
                 elif kw_standalone.value == "key":
                     stream.expect("equals")
@@ -288,9 +383,7 @@ class Parser:
 
                 elif kw_standalone.value == "translate":
                     stream.expect("equals")
-                    standalone_contents["translate"] = stream.expect("string").value[
-                        1:-1
-                    ]
+                    standalone_contents["translate"] = self.get_string(stream)
 
                 elif kw_standalone.value == "with":
                     stream.expect("equals")
@@ -298,7 +391,7 @@ class Parser:
 
                 elif kw_standalone.value == "nbt":
                     stream.expect("equals")
-                    standalone_contents["nbt"] = stream.expect("string").value[1:-1]
+                    standalone_contents["nbt"] = self.get_string(stream)
 
                 elif kw_standalone.value in {"block", "entity", "storage"}:
                     stream.expect("equals")
@@ -308,9 +401,9 @@ class Parser:
 
                 elif kw_standalone.value == "score":
                     stream.expect("equals")
-                    player_name = stream.expect("string").value[1:-1]
+                    player_name = self.get_string(stream)
                     stream.expect("arrow")
-                    objective_name = stream.expect("string").value[1:-1]
+                    objective_name = self.get_string(stream)
                     standalone_contents["score"] = {
                         "name": player_name,
                         "objective": objective_name,
@@ -328,9 +421,7 @@ class Parser:
                 raise InvalidSyntax("Unknown standalone element input")
 
         if stream.get("comma"):
-            standalone_contents = (
-                standalone_contents | self.parse_non_template_standalone(stream)
-            )
+            standalone_contents |= self.parse_non_template_standalone(stream)
 
         return standalone_contents
 
@@ -378,7 +469,7 @@ class Parser:
         args = []
         if stream.get("comma"):
             # 1:-1 to remove quotes
-            args.append(stream.expect("string").value[1:-1])
+            args.append(self.get_string(stream))
             # Add the previous argument to the list
             args.extend(self.parse_template_args(stream))
 
@@ -429,15 +520,20 @@ class Parser:
                     # Just "with" for now
                     modifier_contents[kw_json.value] = json
 
-            elif kw_scope:
-                # Just "hover_text" for now
+            elif kw_scope:  # Just "hover_text" for now
+                self.error_if_scope(stream)
                 stream.expect("equals")
                 stream.expect(("scope", "<"))
                 with stream.provide(scoped=True):
+                    # Remove the existing parent for a bit
+                    q = self.parent
+                    self.parent = ""
                     modifier_contents["hoverEvent"] = {
                         "action": "show_text",
                         "contents": self.internal_parse(stream),
                     }
+                    # Put the old parent back (even though we shouldn't need it for this parse)
+                    self.parent = q
                 stream.expect(("scope", ">"))
 
             elif kw_color or hex_code:
@@ -459,9 +555,9 @@ class Parser:
                     link = url.value
                 elif kw_link:
                     stream.expect("equals")
-                    link = stream.expect("string").value[1:-1]
+                    link = self.get_string(stream)
                     if not link.startswith("http"):
-                        link = "https://" + link
+                        link = f"https://{link}"
                 modifier_contents["clickEvent"] = {"action": "open_url", "value": link}
 
             elif kw_string:
@@ -469,7 +565,7 @@ class Parser:
                 if kw_string.value != "font":
                     self.error_if_scope(stream)
                 stream.expect("equals")
-                string_value = stream.expect("string").value[1:-1]
+                string_value = self.get_string(stream)
                 if kw_string.value == "copy":
                     modifier_contents["clickEvent"] = {
                         "action": "copy_to_clipboard",
@@ -510,10 +606,10 @@ class Parser:
                 if (q := pattern.value) not in self.patterns:
                     raise InvalidSyntax(f"Unknown pattern '{q}'")
 
-                modifier_contents = modifier_contents | self.patterns[q]
+                modifier_contents |= self.patterns[q]
 
             if stream.get("comma"):
-                modifier_contents = modifier_contents | self.parse_modifiers(stream)
+                modifier_contents |= self.parse_modifiers(stream)
 
         return modifier_contents
 
@@ -529,7 +625,7 @@ class Parser:
         return cleaned_ast
 
     def convert_ast_to_json(self, ast: list) -> list:
-        output = [""]
+        output = [self.parent or ""]
         for item in ast:
             if isinstance(item, str):
                 output.append({"text": item})
